import datetime

import pytest
from django.conf import settings
from django.core import mail
from django.test import override_settings
from orgreg_client.models import OrgUnit, OrgUnitList

from greg.models import (
    OrganizationalUnit,
    OuIdentifier,
    Person,
    Role,
    RoleType,
    Sponsor,
    SponsorOrganizationalUnit,
)
from greg.tasks import import_from_orgreg, notify_sponsors_roles_ending
from gregui.models import EmailTemplate


@pytest.fixture
def role_end_reminder_template():
    et = EmailTemplate.objects.create(
        template_key=EmailTemplate.EmailType.ROLE_END_REMINDER,
        subject="role end reminder",
        body="""Dette er en automatisk generert melding fra gjesteregistreringstjenesten.
Du administrerer {{ num_roles }} gjesteroller som snart utløper.
Dersom det er ønskelig å forlenge rollene kan dette gjøres i webgrensesnittet.

This message has been automatically generated by the guest registration system.
You can maintan {{ num_roles }} guest roles about to expire.
To extend the guest role please use the web interface.""",
    )
    return EmailTemplate.objects.get(id=et.id)


@pytest.fixture
def org_tree_with_sponsors(
    person_foo: Person, role_type_test_guest: RoleType, role_type_foo: RoleType
):
    # Create all units
    a = OrganizationalUnit.objects.create()
    b = OrganizationalUnit.objects.create(parent=a)
    c = OrganizationalUnit.objects.create(parent=b)

    # Create all sponsors
    d = Sponsor.objects.create(
        feide_id="d@example.org",
        first_name="D",
        last_name="Guy",
        work_email="d_guy@example.com",
    )
    e = Sponsor.objects.create(
        feide_id="e@example.org",
        first_name="E",
        last_name="Guy",
        work_email="e_guy@example.com",
    )
    f = Sponsor.objects.create(
        feide_id="f@example.org",
        first_name="F",
        last_name="Guy",
        work_email="f_guy@example.com",
    )
    g = Sponsor.objects.create(
        feide_id="g@example.org",
        first_name="G",
        last_name="Guy",
        work_email="g_guy@example.com",
    )
    # Connect them to units
    SponsorOrganizationalUnit.objects.create(
        sponsor=d, organizational_unit=a, hierarchical_access=True
    )
    SponsorOrganizationalUnit.objects.create(
        sponsor=e, organizational_unit=b, hierarchical_access=True
    )

    # Create person with role on unit c and b
    Role.objects.create(
        person=person_foo,
        type=role_type_test_guest,
        start_date=datetime.date.today(),
        end_date=datetime.date.today() + datetime.timedelta(days=5),
        sponsor=f,
        orgunit=c,
    )
    Role.objects.create(
        person=person_foo,
        type=role_type_foo,
        start_date=datetime.date.today(),
        end_date=datetime.date.today() + datetime.timedelta(days=5),
        sponsor=e,
        orgunit=b,
    )

    return a, b, c, d, e, f, g


@pytest.mark.django_db
def test_notify_sponsors_roles_ending(
    role_end_reminder_template, role_person_foo2, sponsor_org_unit
):
    mail.outbox = []
    task_ids = notify_sponsors_roles_ending()
    assert len(task_ids) == 1


@pytest.mark.django_db
def test_sponsor_get_email(role_end_reminder_template, org_tree_with_sponsors):
    mail.outbox = []
    task_ids = notify_sponsors_roles_ending()
    assert len(task_ids) == 1


@pytest.mark.django_db
def test_only_close_sponsors_get_email(
    role_end_reminder_template, org_tree_with_sponsors
):
    _, _, c, _, _, f, g = org_tree_with_sponsors
    SponsorOrganizationalUnit.objects.create(
        sponsor=f, organizational_unit=c, hierarchical_access=True
    )
    SponsorOrganizationalUnit.objects.create(
        sponsor=g, organizational_unit=c, hierarchical_access=True
    )
    mail.outbox = []
    task_ids = notify_sponsors_roles_ending()
    assert len(task_ids) == 3


@pytest.fixture
def old_unit():
    ou = OrganizationalUnit.objects.create(name_nb="a", name_en="b")
    return OrganizationalUnit.objects.get(id=ou.id)


@pytest.fixture
def old_identifier(unit_foo):
    ouid = OuIdentifier.objects.create(
        orgunit=unit_foo,
        name=settings.ORGREG_NAME,
        source=settings.ORGREG_SOURCE,
        value="4",
    )
    return OuIdentifier.objects.get(id=ouid.id)


@pytest.fixture
def orgreg_response():
    orgreg_ous = OrgUnitList(
        __root__=[
            OrgUnit(
                ou_id=1,
                valid_from=datetime.date(year=2020, month=2, day=4),
                external_keys=[],
                name={"nob": "foo"},
                acronym={"nob": "foo_acronym"},
            ),
            OrgUnit(
                ou_id=3,
                valid_from=datetime.date(year=2020, month=2, day=6),
                valid_to=datetime.date(year=2020, month=2, day=7),
                parent=2,
                name={"nob": "bar"},
                external_keys=[],
                acronym={"nob": "foo_acronym3"},
            ),
            OrgUnit(
                ou_id=2,
                valid_from=datetime.date(year=2020, month=2, day=5),
                parent=1,
                name={"eng": "baz"},
                external_keys=[],
                acronym={"nob": "foo_acronym2"},
            ),
        ]
    )
    return orgreg_ous


@pytest.mark.django_db
def test_command_ou_init(requests_mock, old_identifier, orgreg_response):
    requests_mock.get("https://example.com/orgreg/ou/", text=orgreg_response.json())
    assert OrganizationalUnit.objects.all().count() == 1
    import_from_orgreg()

    # Ensure all three new units are imported
    assert OrganizationalUnit.objects.all().count() == 4

    # Ensure tree is built correctly
    bottom = OrganizationalUnit.objects.get(
        identifiers__source="orgreg", identifiers__value="3"
    )
    assert bottom.name_nb == "bar"
    assert bottom.parent.identifiers.get().value == "2"
    assert bottom.parent.name_en == "baz"
    assert bottom.parent.parent.identifiers.get().value == "1"
    assert bottom.parent.parent.name_nb == "foo"

    # Ensure unknown org is marked deleted
    old_unit = OrganizationalUnit.objects.get(
        identifiers__source="orgreg", identifiers__value="4"
    )
    assert old_unit.deleted


@pytest.mark.django_db
def test_run_twice(requests_mock, orgreg_response):
    requests_mock.get("https://example.com/orgreg/ou/", text=orgreg_response.json())

    assert OrganizationalUnit.objects.all().count() == 0

    import_from_orgreg()
    assert OrganizationalUnit.objects.all().count() == 3

    import_from_orgreg()
    assert OrganizationalUnit.objects.all().count() == 3


@pytest.mark.django_db
@override_settings(ORGREG_ACRONYMS=["nob"])
def test_import_acronym(requests_mock, orgreg_response):
    requests_mock.get("https://example.com/orgreg/ou/", text=orgreg_response.json())

    import_from_orgreg()
    assert OrganizationalUnit.objects.all().count() == 3

    assert (
        OuIdentifier.objects.get(orgunit__id=1, name="acronym_nob").value
        == "foo_acronym"
    )
    assert (
        OuIdentifier.objects.get(orgunit__id=2, name="acronym_nob").value
        == "foo_acronym2"
    )
    assert (
        OuIdentifier.objects.get(orgunit__id=3, name="acronym_nob").value
        == "foo_acronym3"
    )
